# objects for level specification

import copy
import string

# really just checks for clockwise, we insist that our travels are
# clockwise on the outside of a shape, a counterclockwise turn must then
# be turning away from the interior, thus concave
def is_convex(dir1, dir2):
    clockwise = (-dir1[1],dir1[0])
    if dir2[0] < 0 and not clockwise[0] < 0: return False
    if dir2[0] == 0 and not clockwise[0] == 0: return False
    if dir2[0] > 0 and not clockwise[0] > 0: return False

    if dir2[1] < 0 and not clockwise[1] < 0: return False
    if dir2[1] == 0 and not clockwise[1] == 0: return False
    if dir2[1] > 0 and not clockwise[1] > 0: return False

    return True

def points_to_moves(points):
    moves = []
    horizontal = True
    for i in range(1,len(points)):
        prev_p = points[i-1]
        p = points[i]

        if horizontal:
            moves.append(p[0]-prev_p[0])
        else:
            moves.append(p[1]-prev_p[1])

        horizontal = not horizontal

    return moves

class Block:
    def __init__(self, pos, lines):
        # walk around shape to find all points
        # must be going CLOCKWISE around the OUTSIDE
        cur_pos = pos
        self.points = []
        self.convex = []
        self.extent = (cur_pos,cur_pos) # bounding box, ((minx,miny), (maxx,maxy))
        horiz = True   # first line is horizontal
        # last line (previous for the first point) is vertical
        prev_dir = (0, lines[-1])
        for L in lines:
            self.points.append(cur_pos)
            if horiz:
                cur_pos = (cur_pos[0] + L, cur_pos[1])
                cur_dir = (L,0)
            else:
                cur_pos = (cur_pos[0], cur_pos[1] + L)
                cur_dir = (0,L)
            horiz = not horiz
            self.convex.append(is_convex(prev_dir, cur_dir))
            prev_dir = cur_dir
            self.extent = ((min(self.extent[0][0], cur_pos[0]),
                            min(self.extent[0][1], cur_pos[1])),
                           (max(self.extent[1][0], cur_pos[0]),
                            max(self.extent[1][1], cur_pos[1])))
        if cur_pos != self.points[0]:
            raise ValueError('shape is not closed')

    def translated_copy(self, delta):
        r = copy.copy(self)
        r.points = copy.copy(self.points)   # we need to modify this list
        for i in range(len(r.points)):
            r.points[i] = (r.points[i][0]+delta[0], r.points[i][1]+delta[1])
        r.extent = ((r.extent[0][0]+delta[0],r.extent[0][1]+delta[1]),
                    (r.extent[1][0]+delta[0],r.extent[1][1]+delta[1]))
        return r

    def dump_to_string(self):
        arr = [self.points[0][0],self.points[0][1]]
        horiz = True
        return ','.join(map(str, arr + points_to_moves(self.points+[self.points[0]]))) + '\n'

def dump_level(filename, blocks):
    f = open(filename, 'w')
    try:
        f.writelines(map(Block.dump_to_string, blocks))
    finally:
        f.close()
    
def load_level(filename):
    f = open(filename, 'r')
    try:
        blocks = []
        for block_str in f.readlines():
            ints = map(int, block_str.strip().split(','))
            blocks.append(Block((ints[0],ints[1]), ints[2:]))
    finally:
        f.close()

    return blocks

# generate a (translated) subset of blocks representing the current screen,
# with clipping (clipped objects are coalesced with border into blocks)
# assumes blocks never overlap, line goes clockwise around outside of block

def clip_world(blocks, world_offset, resolution):
    result = []
    clip_list = [] # potential blocks for clipping

    screen_extent = (world_offset, (world_offset[0]+resolution[0]-1, world_offset[1]+resolution[1]-1)) 
    anti_world_offset = (-world_offset[0], -world_offset[1])

    for b in blocks:
        # note: == is on-screen, but these intersect the border so we don't
        # count it. This allows merging blocks that just touch the border on
        # one side but are otherwise entirely onscreen, as well as ignoring
        # those that just touch the border from the wrong side
        left_onscreen  = b.extent[0][0] > screen_extent[0][0] and\
                         b.extent[0][0] < screen_extent[1][0]
        right_onscreen = b.extent[1][0] > screen_extent[0][0] and\
                         b.extent[1][0] < screen_extent[1][0]
        top_onscreen   = b.extent[0][1] > screen_extent[0][1] and\
                         b.extent[0][1] < screen_extent[1][1]
        bot_onscreen   = b.extent[1][1] > screen_extent[0][1] and\
                         b.extent[1][1] < screen_extent[1][1]

        # TODO: has to be a cheaper way to do this than copying all the
        # onscreen ones; we should be able to just build references, and have
        # the translation happen during fov.index_blocks?
        if left_onscreen and right_onscreen and top_onscreen and bot_onscreen:
            result.append(b.translated_copy(anti_world_offset))
        else:
            if (left_onscreen or right_onscreen or \
                (b.extent[0][0] <= screen_extent[0][0] and b.extent[1][0] >= screen_extent[1][0])) \
                and (top_onscreen or bot_onscreen or \
                (b.extent[0][1] <= screen_extent[0][1] and b.extent[1][1] >= screen_extent[1][1])):
                clip_list.append(b)

    def in_range(p):
        return p[0] > screen_extent[0][0] and p[0] < screen_extent[1][0] and\
               p[1] > screen_extent[0][1] and p[1] < screen_extent[1][1]

    def cross_over(p1,p2):
        # does a (horizontal or vertical) line cross the screen?
        return \
            ((p1[1] > screen_extent[0][1] and p1[1] < screen_extent[1][1]) and\
            ((p1[0] <= screen_extent[0][0] and p2[0] >= screen_extent[1][0]) or\
             (p2[0] <= screen_extent[0][0] and p1[0] >= screen_extent[1][0])))\
            or\
            ((p1[0] > screen_extent[0][0] and p1[0] < screen_extent[1][0]) and\
            ((p1[1] <= screen_extent[0][1] and p2[1] >= screen_extent[1][1]) or\
             (p2[1] <= screen_extent[0][1] and p1[1] >= screen_extent[1][1])))

    def nearest(p):
        if p[0] <= screen_extent[0][0]:
            return (0,p[1]-screen_extent[0][1])
        if p[0] >= screen_extent[1][0]:
            return (resolution[0]-1,p[1]-screen_extent[0][1])
        if p[1] <= screen_extent[0][1]:
            return (p[0]-screen_extent[0][0],0)
        if p[1] >= screen_extent[1][1]:
            return (p[0]-screen_extent[0][0],resolution[1]-1)
        raise RuntimeException("wasn't offscreen?")

    # generate limbs
    limb_list = []
    for b in clip_list:
        first_limb = len(limb_list)
        limb = []
        prev_p = b.points[-1]
        for p in b.points:
            if in_range(p):
                # in range
                if limb == []:
                    if not in_range(prev_p):
                        limb.append(nearest(prev_p))
                limb.append((p[0]-screen_extent[0][0],p[1]-screen_extent[0][1]))
            else:
                # out of range
                if limb != []:
                    # end point!
                    limb.append(nearest(p))
                    limb_list.append(limb)
                    limb = []
                elif cross_over(prev_p,p):
                    # both points are out of range, but the line still crosses
                    # the range
                    limb_list.append([nearest(prev_p), nearest(p)])
            prev_p = p
        if limb != []:
            # if the last point is in range, and the first point is in range,
            # then they are part of the same limb, merge
            if in_range(b.points[0]):
                limb_list[first_limb] = limb + limb_list[first_limb]
            # otherwise this in an end point
            else:
                limb.append(nearest(b.points[0]))
                limb_list.append(limb)

    def key_limb(p):
        if p[1] == 0: return (0,-p[0])  # top is edge 0, right to left
        if p[0] == 0: return (1,p[1])  # left is edge 1, top to bottom
        if p[1] == resolution[1]-1: return (2,p[0]) # bottom is edge 2, left to right
        if p[0] == resolution[0]-1: return (3,-p[1]) # right is edge 3, bottom to top

    limb_roots = []
    i = 0
    for L in limb_list:
        # limb_roots = (first, (edge, edgepos), limb, limb_id)
        limb_roots.append((True, key_limb(L[0]), L, i))
        limb_roots.append((False, key_limb(L[-1]), L, i))
        i += 1

    del limb_list
    limb_roots.sort(key=lambda L: L[1])

    if limb_roots == []:
        # degenerate case with nothing to clip
        result.append(Block((resolution[0]-1,0), [-(resolution[0]-1), resolution[1]-1, resolution[0]-1, -(resolution[1]-1)]))
    else:
        # border blocks loop
        while len(limb_roots) > 0:
            border = []
            first_incoming_idx = map(lambda L: L[0], limb_roots).index(False)
            first_incoming = limb_roots[first_incoming_idx]
            start_id = first_incoming[3]
            del limb_roots[first_incoming_idx]
            # index may wrap when list shortens
            first_incoming_idx %= len(limb_roots)

            border.append(first_incoming[2][-1])
            
            # follow border and limbs until we come back to the starting limb
            last_root = first_incoming
            # next root
            search_idx = first_incoming_idx
            while True:
                next_root = limb_roots[search_idx]
                if not next_root[0]:
                    raise RuntimeError("this loop should always hit outgoing")
                # draw line from last root (outgoing)
                start_point = last_root[1]
                while start_point[0] != next_root[1][0] or start_point > next_root[1]:
                    # last root was on different edge, or later on this edge,
                    # draw corners
                    if start_point[0] == 0:
                        # top ends at upper left
                        border.append((0,0))
                        start_point = (1,0)
                    elif start_point[0] == 1:
                        # left ends at lower left
                        border.append((0,resolution[1]-1))
                        start_point = (2,0)
                    elif start_point[0] == 2:
                        # bottom ends at lower right
                        border.append((resolution[0]-1,resolution[1]-1)) 
                        start_point = (3,-(resolution[1]-1))
                    elif start_point[0] == 3:
                        # right ends at upper left
                        border.append((resolution[0]-1,0))
                        start_point = (0,-(resolution[0]-1))

                # draw the limb itself
                border += next_root[2]

                # this outgoing root is done
                del limb_roots[search_idx]
                # index may wrap but we don't need it anymore

                # have we now made it back to the beginning?
                if next_root[3] == start_id: break

                # find corresponding incoming root
                search_idx = map(lambda L: L[3], limb_roots).index(next_root[3])

                last_root = limb_roots[search_idx]

                # we've done the whole limb, so the incoming root is done, too
                del limb_roots[search_idx]
                # index may wrap when list shortens
                search_idx %= len(limb_roots)

                # search_idx now points to the next limb root to follow

            if border[0][1] != border[1][1]:
                # need to rotate points if first edge is vertical, throw out
                # first point and duplicate second point at end
                border = border[1:] + [border[2]]

            # show me your moves!
            moves = points_to_moves(border)
            result.append(Block(border[0], moves))

    return result
